/* 
   txt2snd - tab-separated-value to soundfile converter

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <math.h>
#include <sndfile.h>
#include <string.h>
#include <inttypes.h>
#include <config.h>

static void usage (int status);
char *program_name;

static struct option const long_options[] =
{
  {"samplerate", required_argument, 0, 's'},
  {"Vmult", required_argument, 0, 'm'},
  {"Voffset", required_argument, 0, 'o'},
  {"verbose", no_argument, 0, 'v'},
  {"debuglevel", required_argument, 0, 'd'},
  {"version", no_argument, 0, 'V'},
  {"aliki", no_argument, 0, 'A'},
  {"sndfmt", required_argument, 0, 'F'},
};

int o_samplerate = 48000;
float o_mult = 1.0;
float o_off = 0.0;
int o_verbose = 0;
int o_debug = 0;
int o_filetype = 1;
int o_sndfmt = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);

static int decode_switches (int argc, char **argv) {
  int c;

  while ((c = getopt_long (argc, argv, 
			   "o:"	/* offset */
			   "m:"	/* offset */
			   "s:"	/* samplerate */
			   "v"	/* verbose */
			   "d:"	/* debug-level */
			   "h"	/* help */
			   "A"	/* aliki */
			   "F:"	/* sndfile-format */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'm':
	  o_mult = atof(optarg);
	  break;
	case 'o':
	  o_off = atof(optarg);
	  break;
	case 's':
	  o_samplerate = atoi(optarg);
	  break;
	case 'd':
	  o_debug = atoi(optarg);
	  break;
	case 'F':
	  o_sndfmt = atoi(optarg);
	  break;
	case 'A':
	  o_filetype = 0;
	  break;
	case 'v':
	  o_verbose = 1;
	  break;
	case 'V':
	  printf ("%s %s\n", program_name, VERSION);
	  exit (0);
	case 'h':
	  usage (0);
	default:
	  usage (EXIT_FAILURE);
	}
    }
  return optind;
}


//aliki

#define HDRSIZE 256

void * my_open_aliki(char *fn) {
  char p[HDRSIZE];
  FILE	*aldfile;
  if ((aldfile = fopen (fn, "w")) == 0) {
    printf ("Error: Not able to open output file '%s'\n", fn);
    exit (1);
  }

  strcpy (p, "aliki");
  p [6] = p [7] = 0;
  *(uint32_t *)(p +  8) = 2; //_vers
  *(uint32_t *)(p + 12) = 1; // _type;
  *(uint32_t *)(p + 16) = o_samplerate; //_rate_n;
  *(uint32_t *)(p + 20) = 1; //_rate_d;
  *(uint32_t *)(p + 24) = 486239; //_n_fram;
  *(uint32_t *)(p + 28) = 1; // _n_sect;
  *(uint32_t *)(p + 32) = 0; // _tref_i;
  *(uint32_t *)(p + 36) = 0; // _tref_n;
  *(uint32_t *)(p + 40) = 1; // _tref_d;
  *(uint32_t *)(p + 44) = 0; // _bits;

  memset (p + 48, 0, HDRSIZE - 48);
  if (fwrite (p, 1, HDRSIZE, aldfile) != HDRSIZE) {
    printf ("Error: Not able to write aliki header to '%s'\n", fn);
    fclose (aldfile);
    exit(1);
  }

  return ((void*) aldfile);
}

int my_write_aliki(void *d, float val) {
  //fwrite(&val, sizeof (float), 1, (FILE*) d); // _type == 2
  return(fwrite(&val, sizeof (float), 1, (FILE*) d));
}

void my_close_aliki(void *d) {
  fclose((FILE*) d);
}

// soundfile 
void * my_open_sf(char *fn) {
  SNDFILE *outfile ;

  SF_INFO sfinfo ;

  sfinfo.samplerate = o_samplerate;
  sfinfo.channels = 1;
  sfinfo.frames = 0;
  sfinfo.format = o_sndfmt;

  if ((outfile = sf_open (fn, SFM_WRITE, &sfinfo)) == NULL) {
    printf ("Error: Not able to open output file '%s'\n", fn);
    exit (1) ;
  }

#if 1
  sf_command (outfile, SFC_SET_UPDATE_HEADER_AUTO, NULL, SF_TRUE) ;
  sf_command (outfile, SFC_SET_CLIPPING, NULL, SF_TRUE) ;
#endif

  return ((void*)outfile);
}

int my_write_sf(void *d, float val) {
  return (sf_writef_float ((SNDFILE*) d, &val, 1));
}

void my_close_sf(void *d) {
  sf_close((SNDFILE*) d);
}

//


int main (int argc, char **argv) {
  program_name = argv[0];

  int i;
  i = decode_switches (argc, argv);
  if (i != argc-1) usage(1);

  void * outfile;
  void (*p_close)(void*);
  void *(*p_open)(char*);
  int  (*p_write)(void*, float);

  if (o_filetype>0) {
    p_close = &my_close_sf;
    p_open = &my_open_sf;
    p_write = &my_write_sf;
  } else {
    p_close = &my_close_aliki;
    p_open = &my_open_aliki;
    p_write = &my_write_aliki;
  }

  outfile = p_open (argv[i]);

  float time,in,out;
  int  idx, gndcnt;
  int duptot, dupcnt;
  int mistot;
  float min,max,gnd,dupbuf;
  int laststamp = -1;
  int timestamp = 0;
  int wa_diff = 0;
  dupbuf=min=max=0.0;
  gnd=0.0;
  gndcnt=0;
  mistot=duptot=dupcnt=0;

  do {  
  /* format (tab-seprarated)
   * idx     time[ms]        [input/ignored] [output]
   * 11      8.916652e-05    1.846313e-02    -7.477363e-01
   */
    if (4!=(fscanf(stdin,"%i\t%f\t%f\t%f\n", &idx,&time,&in,&out)))break;
    if (o_debug&8)    
      printf("read idx:%i t=%f i=%f o=%f\n",idx,time*48000.0, in, out);

    timestamp= (int) rint(time * (double)o_samplerate);

    float val = o_off + out * o_mult;

    if (val< min) min=val;
    if (val> max) max=val;

    if (timestamp > laststamp) {
      int i;

      if (dupcnt == 1 && (timestamp-laststamp) == 2) {
	p_write (outfile, dupbuf);
	dupcnt--;
	gnd+=dupbuf; gndcnt++; laststamp++;
	if (o_verbose) 
	  printf("---- worked around time rounding.\n");
	if (o_debug&4 ) 
	  printf("buf: %7i | %7i - 1 frame(s) - [%f]\n",idx-1,laststamp, dupbuf);
      }

      if ((timestamp-laststamp) == 2 && wa_diff < 5) {
	if (o_verbose) 
	  printf("-++- worked around time rounding.\n");
        wa_diff++; 
	timestamp--;
      } else wa_diff=0;

      if (dupcnt > 0) { 
        duptot++;
	if (o_debug&2) 
	  printf("!--- ignored %i dup. for sample %i\n",dupcnt, laststamp);
      }

      if (o_debug&4 ) 
	printf("save %7i | %7i - %i frame(s) - [%f]\n",idx,timestamp, (timestamp-laststamp), val);
      if (o_verbose && (timestamp-laststamp)>1)
	printf("!!!! inserted %i frame(s) @%i [%f]\n",(timestamp-laststamp)-1,timestamp,val);

      for (i=0;i<(timestamp-laststamp);i++) {
	p_write (outfile, val);
	gnd+=val; gndcnt++;
      }

      mistot+=timestamp-laststamp-1;
      laststamp=timestamp;
      dupcnt=0;
    } else {
      dupcnt++;
      dupbuf=val;
      if ((o_debug&1)) 
	printf("dup! %7i | %7i [%f]\n",idx,timestamp, val);
    }
    
  } while(1);

  p_close(outfile);

  printf(" min:%f - max:%f (2/delta: %f) gnd:%f\n",min,max, 2.0/(max-min), (gnd/(double)gndcnt));
  printf(" duplicate/missing sample info: %i/%i\n",duptot,mistot);
  return(0);
}

static void usage (int status) {
  printf ("%s %s - \
tab-seperated-text to sound file converter\n", program_name, VERSION);
  printf ("reads from <stdin> and writes 24bit signed-le wav mono files.\n");
  printf ("Usage: %s [OPTION]... <outfile>...\n", program_name);
  printf ("\
Options:\n\
  -s <int>                   samplerate [48000]\n\
  -o <float>                 GND offset  [0.0]\n\
  -m <float>                 amplitude multiplier [1.0]\n\
  -v, --verbose              print processing information\n\
  -d <int>, --debug <int>    debug level (bits0-3)\n\
  -A, --aliki                write aliki .ald format\n\
  -F <int>, --sndfnt <int>   specify sndfile format\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
  \n\
Input Format:\n\
  4 tabulator separated values per line:\n\
  <idx/int> <time|ms/float> <ignored/float> <value/float>\n\
11      8.916652e-05    1.846313e-02    -7.477363e-01\n\
\n\
");
#define DA(val) val, val
  printf ("Common sndfile formats (see sndfile.h):\n");
  printf (" WAV  - 24 signed LE : 0x%x = %i (default)\n", DA(SF_FORMAT_WAV|SF_FORMAT_PCM_24));
  printf (" WAV  - 16 signed LE : 0x%x = %i\n", DA(SF_FORMAT_WAV|SF_FORMAT_PCM_16));
  printf (" WAV - 8 unsigned LE : 0x%x = %i\n", DA(SF_FORMAT_WAV|SF_FORMAT_PCM_U8));
  printf (" AIFF - 16 signed LE : 0x%x = %i\n", DA(SF_FORMAT_AIFF|SF_FORMAT_PCM_16));
  printf (" AIFF - 24 signed BE : 0x%x = %i\n", DA(SF_FORMAT_AIFF|SF_FORMAT_PCM_24|SF_ENDIAN_BIG));

  exit (status);
}

/* vi:set ts=8 sts=2 sw=2: */

